@@ -6,8 +6,10 @@ from pasc.model import RequestBatch, ResultBatch, ClientRequest, ClientResult
 
 SCRAPESTACK_URL = "https://api.scrapestack.com/scrape"
 
+# todo: add parsing/model validation
 
-async def async_request(
+
+async def __async_request(
     session: ClientSession,
     request: ClientRequest,
     secret: str,
@@ -28,8 +30,12 @@ async def async_request(
             },
         ) as response:
             data = await response.read()
-            status = int(str(response.status))
+            status = int(str(response.status))  # todo: fix patch later
             duration = perf_counter() - start
+            if request.parser is not None:
+                data = request.parser(data)
+            if request.validator is not None:
+                data = request.validator.model_validate(data)
             return ClientResult(
                 target=request.target,
                 method=request.method,
@@ -40,6 +46,7 @@ async def async_request(
                 status=int(status),
                 duration=duration,
                 data=data,
+                group=request.group,
             )
     except TimeoutError:
         duration = perf_counter() - start
@@ -53,16 +60,17 @@ async def async_request(
             status=408,
             duration=duration,
             data=None,
+            group=request.group,
         )
 
 
-async def execute_batch(batch: RequestBatch, secret: str) -> ResultBatch:
+async def __async_execute_batch(batch: RequestBatch, secret: str) -> ResultBatch:
     global_start = perf_counter()
     async with ClientSession() as session:
         tasks = list()
         for request in batch.requests:
             task = asyncio.ensure_future(
-                async_request(
+                __async_request(
                     session,
                     request,
                     secret,
@@ -76,37 +84,63 @@ async def execute_batch(batch: RequestBatch, secret: str) -> ResultBatch:
     )
 
 
-def execute(
+# todo: implement
+# todo: docstrings
+def execute_one(
+    request: ClientRequest,
+    secret: str,
+) -> ClientResult:
+    raise NotImplementedError()
+
+
+# todo: docstring example
+def execute_batch(
     batch: RequestBatch,
     secret: str,
 ) -> ResultBatch:
     """
-    _summary_
+    Execute a batch of requests outside a running event loop (eg: jupyter, uvicorn).
 
     Args:
-        batch (RequestBatch): _description_
-        secret (str): Scrapestack API/access key.
-        verbose (bool, optional): Show progress bar. Defaults to False.
+        batch (RequestBatch): a batch of requests
+        secret (str): scrapestack api key
 
     Returns:
-        ResultBatch: _description_
+        ResultBatch: a batch of results
     """
-    return asyncio.run(execute_batch(batch, secret))
+    return asyncio.run(__async_execute_batch(batch, secret))
 
 
-def execute_in_event_loop(
+# todo: implement
+# todo: docstring details
+def el_execute_one(
+    request: ClientRequest,
+    secret: str,
+) -> ClientResult:
+    """
+    Execute a single request inside a running event loop (eg: jupyter, uvicorn).
+
+    Args:
+        request (ClientRequest): _description_
+        secret (str): _description_
+
+    Returns:
+        ClientResult: _description_
+
+    Usage:
+
+    ```python
+    from pasc import proxy, ClientRequest
+    request = ClientRequest(target="https://en.wikipedia.org", method="GET")
+    result = await proxy.el_execute_one(request=request, secret="<your API key>")
+    ```
+    """
+    raise NotImplementedError()
+
+
+# todo: docstrings
+def el_execute_batch(
     batch: RequestBatch,
     secret: str,
 ) -> ResultBatch:
-    """
-    _summary_
-
-    Args:
-        batch (RequestBatch): _description_
-        secret (str): Scrapestack API/access key.
-        verbose (bool, optional): Show progress bar. Defaults to False.
-
-    Returns:
-        ResultBatch: _description_
-    """
-    return execute_batch(batch, secret)
+    return __async_execute_batch(batch, secret)
